using System;

namespace Params
{
    //simple class
    class MyClass {
        public int a, b;  //some integer fields

        //constructor
        public MyClass(int a, int b) {
            this.a = a;
            this.b = b;
        }
        
        //Value Parameters
        /// <summary>
        /// Increases value of the parameter by 2.
        /// </summary>
        /// <param name="param">Input only parameter</param>
        /// <returns>Value of the parameter increased by 2</returns>
        public int ValueMethod(int param){
            param = param + 2;  //increase local copy of param by 2
            return param;
        }

        //Reference Parameters
        /// <summary>
        /// Sets the passed variable to the field 'a' and returns the old value of the passed variable.
        /// </summary>
        /// <param name="param">Input/output parameter</param>
        /// <returns>Old value of the parameter</returns>
        public int ReferenceMethod(ref int param){
            int c = param;  //store the old value of param
            param = a;      //set new value to the variable passed as param
            return c;       //return the old valut of the variable passed as param
        }

        //Output Parameters
        /// <summary>
        /// Stores to the parameter variable the sum of 'a' and 'b' fields. The same is returned.
        /// </summary>
        /// <param name="param">Output only parameter</param>
        /// <returns>The same value as the one returned in the parameter</returns>
        public int OutputMethod(out int param){
            //int c = param;  //error, param is unasigned
            param = a + b;    //set new value to the variable passed as param
            return param;     
        }

        //Parameter Arrays
        /// <summary>
        /// Sets the 'a' field to value of the first parameter and 'b' to value of the second parameter.
        /// </summary>
        /// <param name="param">Variable number of input parameters, only first two are handled</param>
        public void ParamMethod(params int[] param){
            a = param[0];     //get value of the first parameter
            b = param[1];     //get value of the second parameter
        }

        //simple output of 'a' and 'b' to the console 
        public void Print() {
            Console.WriteLine("a = {0}, b = {1}", a, b);
        }

        //Reference type by value
        /// <summary>
        /// Increases the 'a' field of the variable passed as the parameter 'target'. Further does some 
        /// changes on the local copy, but these don't affect the original variable.
        /// </summary>
        /// <param name="target">A reference type variable passed by value</param>
        public static void NewClass(MyClass target){
            MyClass nClass = new MyClass(1, 2);
            //it affects the data of original target
            //the next step has no effect to it
            target.a += 1; 

            //the copy of target points to memory of nClass
            //the original target points to the same place as before
            target = nClass;

            //the original data of target is not affacted
            target.b += 2;
        }

        //Reference type by reference
        /// <summary>
        /// Increases the 'a' field of the variable passed as the parameter 'target'. Futher sets the original 
        /// pointer to a new place (so the previous change of 'a' has no effect) and increases the new value of
        /// 'b' by 2. 
        /// </summary>
        /// <param name="target">A reference type variable passed by reference</param>
        public static void NewClass(ref MyClass target){
            MyClass nClass = new MyClass(1, 2);
            //it affects the data of original target, 
            //but in the next step it will be overwritten
            target.a += 1;

            //the original target points to memory of nClass
            target = nClass;

            //data of original target is affected
            target.b += 2;
        }

    }

    class Program
    {
        static void Main(string[] args)
        {
            int i, j, k;
            MyClass cl = new MyClass(5, 7);

            i = 10;
            j = 15;
            k = 0;
            Console.WriteLine("Original values:");
            Console.WriteLine("i = {0}, j = {1}, k = {2}", i, j, k);

            //k becomes 10+2 = 12, i is still 10
            k = cl.ValueMethod(i);
            Console.WriteLine("\nValue Parameters:");
            Console.WriteLine("i = {0}, k = {1}", i, k);

            //k becomes 10 (value of i before cl.ReferenceMethod call), i becomes 5 (value of cl.a)
            k = cl.ReferenceMethod(ref i);
            Console.WriteLine("\nReference Parameters:");
            Console.WriteLine("i = {0}, k = {1}", i, k);

            //k and i becomes 12 ((a = 5) + (b = 7) = 12)
            k = cl.OutputMethod(out i);
            Console.WriteLine("\nOutput Parameters:");
            Console.WriteLine("i = {0}, k = {1}", i, k);

            Console.WriteLine("\nOriginal values of class cl:");
            cl.Print();
            // i and j are unchanged
            // cl.a becomes 12 (value of i), cl.b becomes 15 (value of j)
            Console.WriteLine("\nParameter Array:");
            cl.ParamMethod(i, j);
            cl.Print();


            Console.WriteLine("\nMyClass.NewClass by value:");

            cl.Print();
            MyClass.NewClass(cl);
            cl.Print();

            Console.WriteLine("\nMyClass.NewClass by reference:");

            cl.Print();
            MyClass.NewClass(ref cl);
            cl.Print();
        }
    }
}
